#include "tmesh.h"

#include <fstream>
#include <iostream>
#include <float.h>
#include <math.h>

TMesh::TMesh() : verts(0), vertsN(0), tris(0), trisN(0), 
  colors(0),  normals(0), texture(0), sts(0) {};

void TMesh::SetBaseRectangle(V3 center, float dx, float dz) {

  vertsN = 4;
  trisN = 2;

  verts = new V3[vertsN];
  colors = new V3[vertsN];
  normals = new V3[vertsN];
  tris = new unsigned int[trisN*3];

  verts[0] = center + V3(-dx/2.0f, 0.0f, -dz/2.0f);
  verts[1] = center + V3(-dx/2.0f, 0.0f, dz/2.0f);
  verts[2] = center + V3(dx/2.0f, 0.0f, dz/2.0f);
  verts[3] = center + V3(dx/2.0f, 0.0f, -dz/2.0f);

  normals[0] = 
  normals[1] = 
  normals[2] = 
  normals[3] = V3(0.0f, 1.0f, 0.0f);

  colors[0] = V3(0.0f, 0.0f, 0.0f);
  colors[1] = V3(0.0f, 0.0f, 0.0f);
  colors[2] = V3(1.0f, 1.0f, 0.0f);
  colors[3] = V3(1.0f, 1.0f, 0.0f);

  tris[0] = 0;
  tris[1] = 1;
  tris[2] = 2;
  tris[3] = 2;
  tris[4] = 3;
  tris[5] = 0;

  sts = new float[2*vertsN];
  sts[0] = 0.0f;
  sts[1] = 0.0f;
  sts[2] = 0.0f;
  sts[3] = 1.0f;
  sts[4] = 1.0f;
  sts[5] = 1.0f;
  sts[6] = 1.0f;
  sts[7] = 0.0f;

  texture = new FrameBuffer(0, 0, 1024, 1024);
  texture->MakeChecker(64, 0xFF000000, 0xFFFFFFFF);

}

void TMesh::Load(char *fname) {

  ifstream ifs(fname, ios::binary);
  if (ifs.fail()) {
    cerr << "INFO: cannot open file: " << fname << endl;
    return;
  }

  ifs.read((char*)&vertsN, sizeof(int));
  char yn;
  ifs.read(&yn, 1); // always xyz
  if (yn != 'y') {
    cerr << "INTERNAL ERROR: there should always be vertex xyz data" << endl;
    return;
  }
  if (verts)
    delete verts;
  verts = new V3[vertsN];

  ifs.read(&yn, 1); // colors 3 floats
  if (colors)
    delete colors;
  colors = 0;
  if (yn == 'y') {
    colors = new V3[vertsN];
  }

  ifs.read(&yn, 1); // normals 3 floats
  if (normals)
    delete normals;
  normals = 0;
  if (yn == 'y') {
    normals = new V3[vertsN];
  }

  ifs.read(&yn, 1); // texture coordinates 2 floats
  float *tcs = 0; // don't have texture coordinates for now
  if (tcs)
    delete tcs;
  tcs = 0;
  if (yn == 'y') {
    tcs = new float[vertsN*2];
  }

  ifs.read((char*)verts, vertsN*3*sizeof(float)); // load verts

  if (colors) {
    ifs.read((char*)colors, vertsN*3*sizeof(float)); // load colors
  }

  if (normals)
    ifs.read((char*)normals, vertsN*3*sizeof(float)); // load normals

  if (tcs)
    ifs.read((char*)tcs, vertsN*2*sizeof(float)); // load texture coordinates

  ifs.read((char*)&trisN, sizeof(int));
  if (tris)
    delete tris;
  tris = new unsigned int[trisN*3];
  ifs.read((char*)tris, trisN*3*sizeof(unsigned int)); // read tiangles

  ifs.close();

  cerr << "INFO: loaded " << vertsN << " verts, " << trisN << " tris from " << endl << "      " << fname << endl;
  cerr << "      xyz " << ((colors) ? "rgb " : "") << ((normals) ? "nxnynz " : "") << ((tcs) ? "tcstct " : "") << endl;


}

void TMesh::RenderWF(FrameBuffer *fb, PPC *ppc, 
                     unsigned int wfc) {

  V3 *projVerts = new V3[vertsN];
  for (int i = 0; i < vertsN; i++) {
    ppc->Project(verts[i], projVerts[i]);
  }

  for (int i = 0; i < trisN; i++) {
    for (int j = 0; j < 3; j++) {
      int _j = (j+1)%3;
      int vi0  = tris[3*i+j];
      int vi1  = tris[3*i+_j];
      V3 c0, c1;
      if (colors) {
        c0 = colors[vi0];
        c1 = colors[vi1];
      }
      else {
        c0.SetFromColor(wfc);
        c1 = c0;
      }
      fb->DrawSegment2D(projVerts[vi0], projVerts[vi1], c0, c1);
    }
  }

  delete projVerts;

}

void TMesh::Translate(V3 trans) {

  for (int i = 0; i < vertsN; i++) {
    verts[i] = verts[i] + trans;
  }

}

V3 TMesh::GetCenter() {

  V3 center(0.0f, 0.0f, 0.0f);
  for (int i = 0; i < vertsN; i++) {
    center = center + verts[i];
  }
  center = center / (float) vertsN;
  return center;

}

AABB TMesh::GetAABB() {

  AABB ret;
  for (int i = 0; i < vertsN; i++) {
    ret.AddPoint(verts[i]);
  }

  return ret;

}

void TMesh::Scale(float diag) {

  AABB aabb = GetAABB();
  float currDiag = (aabb[0]-aabb[1]).Length();
  float scalef = diag / currDiag;
  V3 c = (aabb[0]+aabb[1]) / 2.0f;
  for (int i = 0; i < vertsN; i++) {
    verts[i] = c + (verts[i]-c)*scalef;
  }

}

void TMesh::SetCenter(V3 newCenter) {

  V3 oldCenter = GetCenter();
  Translate(newCenter-oldCenter);

}

void TMesh::RenderFilled(FrameBuffer *fb, 
  PPC *ppc, V3 lightPoint, FrameBuffer *shadowMap,
  PPC *lightCamera, bool msi) {
  
  V3 *projVerts = new V3[vertsN];
  for (int i = 0; i < vertsN; i++) {
    ppc->Project(verts[i], projVerts[i]);
  }

  for (int i = 0; i < trisN; i++) {
    
    int vi[3]= {tris[3*i+0], tris[3*i+1], tris[3*i+2]};

    int left, right, top, bottom;
    if (!Clip(projVerts, vi, fb->w, fb->h, left, right, top, bottom))
      continue;

    // edge equation setup
    M3x3 eeqs;
    for (int j = 0; j < 3; j++) {
      int _j = (j+1)%3;
      V3 p = projVerts[vi[j]];
      V3 q = projVerts[vi[_j]];
      V3 r = projVerts[vi[(j+2)%3]];
      eeqs[j][0] = q[1]-p[1];
      eeqs[j][1] = p[0]-q[0];
      eeqs[j][2] = -p[1]*eeqs[j][1]-p[0]*eeqs[j][0];
      if (eeqs[j]*V3(r[0], r[1], 1.0f) < 0.0f) {
        eeqs[j] = eeqs[j]*-1.0f;
      }
    }

    // rasterization parameter setup
    M3x3 ssim;
    ssim[0] = V3(projVerts[vi[0]]); ssim[0][2] = 1.0f;
    ssim[1] = V3(projVerts[vi[1]]); ssim[1][2] = 1.0f;
    ssim[2] = V3(projVerts[vi[2]]); ssim[2][2] = 1.0f;
    ssim = ssim.Invert();

    // z
    V3 ABCz = ssim * V3(projVerts[vi[0]][2], 
      projVerts[vi[1]][2], projVerts[vi[2]][2]);
    V3 sABC, tABC;
    if (sts) {
      sABC = ssim * V3(sts[vi[0]*2+0], sts[vi[1]*2+0], sts[vi[2]*2+0]);
      tABC = ssim * V3(sts[vi[0]*2+1], sts[vi[1]*2+1], sts[vi[2]*2+1]);
    }

    V3 shades[3];
    float ka = 0.4f;
    for (int i = 0; i < 3; i++) {
      V3 lightv = (lightPoint - verts[vi[i]]).Normalized();
      shades[i] = lightv.Light(normals[vi[i]], ka, colors[vi[i]], false);
    }

    // colors
    M3x3 ABCcolors = SetABCMatrix(ssim, shades[0], shades[1], shades[2]);
    M3x3 ABCacolors = SetABCMatrix(ssim, colors[vi[0]]*ka, 
      colors[vi[1]]*ka, colors[vi[2]]*ka);
    M3x3 ABCocolors = SetABCMatrix(ssim, colors[vi[0]], 
      colors[vi[1]], colors[vi[2]]);

    // model space interpolation setup
    M3x3 Q = SetMSIQ(verts[vi[0]], verts[vi[1]], verts[vi[2]], ppc);
    V3 denABC = SetDenABC(Q);
    M3x3 colorNumABC;
    colorNumABC[0] = SetNumABC(Q, 
      V3(colors[vi[0]][0], colors[vi[1]][0], colors[vi[2]][0]));
    colorNumABC[1] = SetNumABC(Q, 
      V3(colors[vi[0]][1], colors[vi[1]][1], colors[vi[2]][1]));
    colorNumABC[2] = SetNumABC(Q, 
      V3(colors[vi[0]][2], colors[vi[1]][2], colors[vi[2]][2]));
    V3 stNumABC[2];
    if (sts) {
      stNumABC[0] = SetNumABC(Q, 
        V3(sts[vi[0]*2+0], sts[vi[1]*2+0], sts[vi[2]*2+0]));
      stNumABC[1] = SetNumABC(Q, 
        V3(sts[vi[0]*2+1], sts[vi[1]*2+1], sts[vi[2]*2+1]));
    }

    for (int v = top; v <= bottom; v++) {
      for (int u = left; u <= right; u++) {
        V3 pixv(.5f+(float)u, .5f+(float)v, 1.0f);
        if (
          eeqs[0]*pixv < 0.0f || 
          eeqs[1]*pixv < 0.0f || 
          eeqs[2]*pixv < 0.0f)
          continue;
        float currz = ABCz*pixv;
        if (fb->IsCloserThenSet(u, v, currz)) {
          V3 currColor = ABCocolors*pixv;
          V3 curraColor = ABCacolors*pixv;

          if (lightCamera) {
            V3 projP(pixv[0], pixv[1], currz);
            V3 P = ppc->UnProject(projP); // point at pixel
            V3 lightProjP;
            if (lightCamera->Project(P, lightProjP)) {
              if (shadowMap->IsInShadow(lightProjP)) {
//                currColor = V3(0.0f, 0.0f, 0.0f);
                currColor = curraColor;
              }
            }
          }

          if (msi) {
            // model space interpolation of color
            float currDen = denABC*pixv;
            currColor = V3(
              colorNumABC[0]*pixv, 
              colorNumABC[1]*pixv, 
              colorNumABC[2]*pixv)/currDen;
            if (texture) {
              float currs = (stNumABC[0]*pixv)/currDen;
              float currt = (stNumABC[1]*pixv)/currDen;
              currColor = texture->LookUp(currs, currt);
            }
          }
          else {
            if (texture) {
              float currs = sABC*pixv;
              float currt = tABC*pixv;
              currColor = texture->LookUp(currs, currt);
            }
          }
          fb->Set(u, v, currColor.GetColor());
        }
      }
    }

  }

  delete []projVerts;

}


void TMesh::SetColors(V3 color) {

  if (!colors)
    return;

  for (int i = 0; i < vertsN; i++) {
    colors[i] = color;
  }

}


bool TMesh::Clip(V3 *projVerts, int *vi, int w, int h,
                 int& left, int& right, int &top, int &bottom) {

    // any vertex behind eye
    if (
      projVerts[vi[0]][0] == FLT_MAX ||
      projVerts[vi[1]][0] == FLT_MAX ||
      projVerts[vi[2]][0] == FLT_MAX) {
      return false;
    }
    AABB projaabb;
    projaabb.AddPoint(projVerts[vi[0]]);
    projaabb.AddPoint(projVerts[vi[1]]);
    projaabb.AddPoint(projVerts[vi[2]]);

    // bbox of proj tri outside of image frame
    if (
      projaabb[0][0] > (float) w ||
      projaabb[1][0] < 0.0f ||
      projaabb[0][1] > (float) h ||
      projaabb[1][1] < 0.0f)
      return false;

    // clip bbox to image frame
    if (projaabb[0][0] < 0.0f)
      projaabb[0][0] = 0.0f;
    if (projaabb[1][0] > (float) w)
      projaabb[1][0] = (float) w;
    if (projaabb[0][1] < 0.0f)
      projaabb[0][1] = 0.0f;
    if (projaabb[1][1] > (float) h)
      projaabb[1][1] = (float) h;

    top = (int) floorf(.5f+projaabb[0][1]);
    left = (int) floorf(.5f+projaabb[0][0]);
    bottom = (int) floorf(-.5f+projaabb[1][1]);
    right = (int) floorf(-.5f+projaabb[1][0]);

    return true;
}


M3x3 TMesh::SetABCMatrix(M3x3 ssim, V3 r0, V3 r1, V3 r2) {


  M3x3 ret;
  ret[0] = ssim * V3(r0[0], r1[0], r2[0]);
  ret[1] = ssim * V3(r0[1], r1[1], r2[1]);
  ret[2] = ssim * V3(r0[2], r1[2], r2[2]);

  return ret;

}

M3x3 TMesh::SetMSIQ(V3 V0, V3 V1, V3 V2, PPC *ppc) {

  M3x3 ret;
  ret.SetColumn(0, V0-ppc->C);
  ret.SetColumn(1, V1-ppc->C);
  ret.SetColumn(2, V2-ppc->C);
  M3x3 camM;
  camM.SetColumn(0, ppc->a);
  camM.SetColumn(1, ppc->b);
  camM.SetColumn(2, ppc->c);
  ret = ret.Invert()*camM;

  return ret;

}

V3 TMesh::SetDenABC(M3x3 Q) {

  M3x3 aux;
  aux[0] = V3(1.0f, 1.0f, 1.0f);
  aux[1] = V3(1.0f, 1.0f, 1.0f);
  aux[2] = V3(1.0f, 1.0f, 1.0f);
  V3 ret = (aux*Q)[0];
  return ret;

}


V3 TMesh::SetNumABC(M3x3 Q, V3 r) {

  V3 ret = Q.Transpose()*r;
  return ret;

}